/* arch_form.c */

#include <Pilot.h>
#include <DataPrv.h>

#include "callback.h"

#include "pz.h"
#include "pz_rcp.h"


/*
 * constants
 */

#define ARCH_SIZE_WIDTH 28
#define ARCH_RATIO_WIDTH 34


/*
 * static variables
 */

static DmOpenRef arch_db;
static char arch_db_name[dmDBNameLength];
static UInt arch_db_attr;

static Boolean arch_db_read_only;
static struct db_spec arch_db_spec;

static struct stream_info *stream_list;
static int num_streams;

static ULong arch_db_total_size;
static ULong arch_db_total_uncompressed;


/*
 * archive database and stream list
 */

Boolean open_arch_db(UInt card_no, LocalID db_id)
{
    arch_db = DmOpenDatabase(card_no, db_id, dmModeReadOnly);

    if (arch_db != NULL) {
	DmDatabaseInfo(card_no, db_id, arch_db_name, &arch_db_attr,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

	arch_db_read_only = true;
	arch_db_spec.card_no = card_no;
	arch_db_spec.db_id = db_id;
    }

    return arch_db != NULL;
}

static void close_arch_db(void)
{
    if (arch_db != NULL) {
	DmCloseDatabase(arch_db);
    }

    arch_db = NULL;
}

static void gen_stream_list(void)
{
    int bad, i;

    if (arch_db != NULL) {
	num_streams = count_streams(arch_db, &stream_list, &bad);

	if (bad != 0) {
	    /* zzz */
	}

	DmDatabaseSize(arch_db_spec.card_no, arch_db_spec.db_id,
		NULL, &arch_db_total_size, NULL);

	arch_db_total_uncompressed = 0;
	for (i = 0; i < num_streams; ++i) {
	    arch_db_total_uncompressed += stream_list[i].uncompressed;
	}
    }
}

static void free_stream_list(void)
{
    if (stream_list != NULL) {
	MemPtrFree(stream_list);
    }

    stream_list = NULL;
    num_streams = 0;
}

static void update_stream_list(void)
{
    free_stream_list();
    gen_stream_list();

    set_table_items(num_streams, get_table_top(), num_streams == 1 ? 0 : -1);
}

static struct stream_info *get_selected_stream(void)
{
    int sel;

    sel = get_table_selection();
    if (sel < 0 || sel >= num_streams) {
	return NULL;
    }

    return &stream_list[sel];
}


/*
 * custom draw procedure
 */

static void draw_arch_item(
	VoidPtr table, Word row, Word column, RectanglePtr bounds)
{
    int index;
    struct stream_info *stream;
    VoidHand h;
    VoidPtr p;
    char buf[12];

    CALLBACK_PROLOGUE

    WinEraseRectangle(bounds, 0);

    index = get_table_top() + row;

    if (index < num_streams) {
	stream = &stream_list[index];

	h = DmQueryRecord(arch_db, stream->index);
	p = MemHandleLock(h);
	draw_fit_chars((CharPtr) p + stream->name_offset, 2,
		bounds->extent.x - 2 - ARCH_SIZE_WIDTH * 2 - 2, bounds, false);
	MemHandleUnlock(h);

	print_size(buf, stream->size + STREAM_TRAILER_SIZE);
	draw_chars(buf, 0, bounds->extent.x - ARCH_SIZE_WIDTH * 2, bounds);

	print_size(buf, stream->uncompressed);
	draw_chars(buf, 0, bounds->extent.x - ARCH_SIZE_WIDTH, bounds);
    }

    CALLBACK_EPILOGUE
}

static void draw_total_size(Word id)
{
    RectangleType bounds;
    char buf[12];

    GetObjectBounds(id, &bounds);
    WinEraseRectangle(&bounds, 0);

    print_ratio(buf, arch_db_total_uncompressed, arch_db_total_size);
    draw_chars(buf, 0, 0, &bounds);

    print_size(buf, arch_db_total_size);
    draw_chars(buf, 0, ARCH_RATIO_WIDTH, &bounds);

    print_size(buf, arch_db_total_uncompressed);
    draw_chars(buf, 0, ARCH_RATIO_WIDTH + ARCH_SIZE_WIDTH, &bounds);
}


/*
 * getting information
 */

static void do_info(void)
{
    struct stream_info *stream;
    DatabaseHdrType db_header;
    int err;
    const char *msg = NULL;

    stream = get_selected_stream();
    if (stream == NULL) {
	return;
    }

    if (msg == NULL) {
	err = uncompress_header(arch_db, stream, &db_header);
	if (err != noErr) {
	    msg = "uncompress_header() failed";
	}
    }

    if (msg == NULL) {
	set_db_info(&db_header,
		stream->uncompressed, stream->size + STREAM_TRAILER_SIZE);
	FrmPopupForm(INFO_FORM_ID);
    }

    if (msg != NULL) {
	FrmCustomAlert(ERROR_ALERT_ID, (char *) msg, " ", " ");
    }
}


/*
 * extracting
 */

static const char *do_extract_db(
	DmOpenRef db, struct stream_info *stream,
	DatabaseHdrPtr db_header, UInt card_no)
{
    LocalID db_id;
    Word button;
    FormPtr saved_form, form;
    int err;

    db_id = DmFindDatabase(card_no, db_header->name);
    if (db_id != NULL) {
	button = FrmCustomAlert(
		REPLACE_ALERT_ID, "database", db_header->name, " ");
	if (button != REPLACE_OK_ID) {
	    return NULL;
	}
	DmDeleteDatabase(card_no, db_id);
    }

    form = start_progress(&saved_form, "Extracting...", db_header->name);

    err = uncompress_db(db, stream, card_no, show_progress);

    end_progress(form, saved_form);

    return err == noErr ? NULL : "do_extract_db() failed";
}

static void do_extract(void)
{
    struct stream_info *stream;
    char buf[dmDBNameLength], *buf1 = buf, *buf2 = &buf[dmDBNameLength / 2];
    Word button;
    DatabaseHdrType db_header;
    UInt card_no = DEFAULT_CARD_NO;
    int err;
    const char *msg = NULL;

    stream = get_selected_stream();
    if (stream == NULL) {
	return;
    }

    if (stream->flag != STREAM_FLAG_VALID) {
	print_stream_name(arch_db, stream, buf, dmDBNameLength);
	button = FrmCustomAlert(UNKNOWN_ALERT_ID, buf, " ", " ");
	if (button == UNKNOWN_READ_ID) {
	    open_read_form(arch_db, stream);
	    return;
	} else if (button != UNKNOWN_EXTRACT_ID) {
	    return;
	}
    }

    if (msg == NULL) {
	err = uncompress_header(arch_db, stream, &db_header);
	if (err != noErr) {
	    msg = "uncompress_header() failed";
	}
    }

    if (msg == NULL) {
	db_header.name[dmDBNameLength - 1] = '\0';
	StrIToA(buf1, stream->size + STREAM_TRAILER_SIZE);
	StrIToA(buf2, stream->uncompressed);
	button = FrmCustomAlert(EXTRACT_ALERT_ID, db_header.name, buf1, buf2);
	if (button == EXTRACT_OK_ID) {
	    msg = do_extract_db(arch_db, stream, &db_header, card_no);
	}
    }

    if (msg != NULL) {
	FrmCustomAlert(ERROR_ALERT_ID, (char *) msg, " ", " ");
    }
}


/*
 * deleting
 */

#ifndef UNPZ
static void do_delete(void)
{
    struct stream_info *stream;
    char buf[dmDBNameLength];
    Word button;
    int err;
    const char *msg = NULL;

    if (arch_db_attr & dmHdrAttrReadOnly) {
	msg = "this archive is read only";
    }

    if (num_streams == 0) {
	button = FrmCustomAlert(DELETE_ALERT_ID, "archive", arch_db_name, " ");
	if (button == DELETE_OK_ID) {
	    close_arch_db();
	    err = DmDeleteDatabase(arch_db_spec.card_no, arch_db_spec.db_id);
	    if (err != noErr) {
		msg = "DmDeleteDatabase() failed";
	    }
	}
	goto done;
    }

    stream = get_selected_stream();
    if (stream == NULL) {
	return;
    }

    if (msg == NULL) {
	print_stream_name(arch_db, stream, buf, dmDBNameLength);
	button = FrmCustomAlert(DELETE_ALERT_ID, "compressed item", buf, " ");
	if (button == DELETE_OK_ID) {
	    if (arch_db_read_only) {
		close_arch_db();

		arch_db = DmOpenDatabase(
			arch_db_spec.card_no, arch_db_spec.db_id,
			dmModeReadWrite);
		if (arch_db == NULL) {
		    arch_db = DmOpenDatabase(
			    arch_db_spec.card_no, arch_db_spec.db_id,
			    dmModeReadOnly);
		    msg = "DmOpenDatabase(dmModeReadWrite) failed";
		    goto done;
		} else {
		    arch_db_read_only = false;
		}
	    }

	    err = remove_stream(arch_db, stream->index, stream->num_records);
	    if (err != noErr) {
		msg = "remove_stream() failed";
	    }

	    FrmUpdateForm(prefs.last_form_id, UPDATE_ARCHIVE);
	}
    }

done:
    if (msg != NULL) {
	FrmCustomAlert(ERROR_ALERT_ID, (CharPtr) msg, " ", " ");
    }

    if (arch_db == NULL) {
	FrmGotoForm(DB_FORM_ID);
    }
}
#endif


/*
 * event handlers
 */

static void do_double_tap(void)
{
    struct stream_info *stream;

    stream = get_selected_stream();
    if (stream == NULL) {
	return;
    }

    if (stream->flag == STREAM_FLAG_VALID) {
	do_extract();
    } else {
	open_read_form(arch_db, stream);
    }
}

Boolean handle_arch_form_event(EventPtr event)
{
    Boolean handled = false;

    CALLBACK_PROLOGUE

    switch (event->eType) {
    case frmOpenEvent:
	FrmSetTitle(FrmGetActiveForm(), arch_db_name);
	FrmDrawForm(FrmGetActiveForm());
	init_table(ARCH_TABLE_ID, ARCH_SCROLLBAR_ID, draw_arch_item);
	FrmUpdateForm(prefs.last_form_id, UPDATE_ARCHIVE);
	handled = true;
	break;

    case frmUpdateEvent:
	draw_free_memory(ARCH_FREE_ID);
	if (event->data.frmUpdate.updateCode == UPDATE_ARCHIVE) {
	    update_stream_list();
	    handled = true;
	}
	draw_total_size(ARCH_TOTAL_ID);
	break;

    case frmCloseEvent:
	free_stream_list();
	close_arch_db();
	break;

    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case ARCH_ABOUT_ID:
	    FrmAlert(ABOUT_ALERT_ID);
	    handled = true;
	    break;

	case ARCH_DONE_ID:
	    FrmGotoForm(DB_FORM_ID);
	    handled = true;
	    break;

	case ARCH_INFO_ID:
	    do_info();
	    handled = true;
	    break;

	case ARCH_EXTRACT_ID:
	    do_extract();
	    draw_free_memory(ARCH_FREE_ID);
	    handled = true;
	    break;

#ifndef UNPZ
	case ARCH_DELETE_ID:
	    do_delete();
	    handled = true;
	    break;
#endif
	}
	break;

    case keyDownEvent:
    case sclRepeatEvent:
    case tblExitEvent:
    case tblSelectEvent:
	handled = handle_table_event(event, do_double_tap);
	break;

    default:
	break;
    }

    CALLBACK_EPILOGUE

    return handled;
}
