/* db_info.c */

#include <Pilot.h>
#include <DataPrv.h>
#ifndef UNPZ
#include <ExgMgr.h>
#endif

#include "callback.h"

#include "pz.h"
#include "pz_rcp.h"


/*
 * constants
 */

#ifndef dmHdrAttrHidden
#define dmHdrAttrHidden 0x0100
#endif

#ifndef dmHdrAttrLaunchableData
#define dmHdrAttrLaunchableData 0x0200
#endif

#define dmHdrAttrsShouldNotModify (dmSysOnlyHdrAttrs | dmHdrAttrCopyPrevention)

#ifndef UNPZ
#define BEAM_TARGET_CREATOR 'lnch'
#endif


/*
 * static variables
 */

static const char attrs_on [16] = "o?????LhscrnbDRr";
static const char attrs_off[16] = "---------------d";

static struct db_spec db_spec;
static DatabaseHdrType db_info;
static ULong db_size;
static ULong db_compressed;
static Boolean db_writable;
static char db_attrs[17];


/*
 * database information
 */

Boolean get_db_info(UInt card_no, LocalID db_id)
{
    int err;

    db_spec.card_no = card_no;
    db_spec.db_id = db_id;

    err = DmDatabaseInfo(card_no, db_id,
	    db_info.name, &db_info.attributes, &db_info.version,
	    &db_info.creationDate, &db_info.modificationDate,
	    &db_info.lastBackupDate, &db_info.modificationNumber,
	    &db_info.appInfoID, &db_info.sortInfoID,
	    &db_info.type, &db_info.creator);
    if (err != noErr) {
	return false;
    }

    err = DmDatabaseSize(card_no, db_id, &db_compressed, &db_size, NULL);
    if (err != noErr) {
	return false;
    }

    db_info.recordList.numRecords = db_compressed;
    db_compressed = 0;
    db_writable = (MemLocalIDKind(db_spec.db_id) == memIDHandle);

    return true;
}

void set_db_info(void *db_header, ULong size, ULong compressed)
{
    db_spec.card_no = 0;
    db_spec.db_id = NULL;

    MemMove(&db_info, db_header, sizeof(DatabaseHdrType));
    db_info.name[dmDBNameLength - 1] = '\0';
    db_size = size;
    db_compressed = compressed;
    db_writable = false;
}


/*
 * update information
 */

static void set_field(Word field_id, char *str, int len)
{
    VoidHand h;
    CharPtr p;

    if (len == 0) {
	len = StrLen(str);
    }

    h = MemHandleNew(len + 1);
    if (h != NULL) {
	p = MemHandleLock(h);
	MemMove(p, str, len);
	p[len] = '\0';
	MemHandleUnlock(h);

	FldSetTextHandle(GetObjectPtr(field_id), (Handle) h);
    }
}

static void set_attrs(UInt attr, Boolean update_field)
{
    char buf[9];
    int i;

    StrPrintF(buf, "%x", attr);
    if (StrLen(buf) == 8) {
	MemMove(buf, &buf[4], 4);
	buf[4] = '\0';
    }

    if (update_field) {
	FldFreeMemory(GetObjectPtr(INFO_ATTR_FIELD_ID));
    }
    set_field(INFO_ATTR_FIELD_ID, buf, 0);
    if (update_field) {
	FldDrawField(GetObjectPtr(INFO_ATTR_FIELD_ID));
    }

    for (i = 0; i < 16; ++i) {
	if ((attr & (1 << (15 - i))) != 0) {
	    db_attrs[i] = attrs_on[i];
	} else {
	    db_attrs[i] = attrs_off[i];
	}
    }
    db_attrs[i] = '\0';

    CtlSetLabel(GetObjectPtr(INFO_ATTR_SELECTOR_ID), db_attrs);
}

static void set_date_field(Word field_id, ULong seconds)
{
    DateFormatType date_format;
    TimeFormatType time_format;
    DateTimeType dt;
    char buf[32];

    if (seconds == 0) {
	StrCopy(buf, "-");
    } else {
	date_format = PrefGetPreference(prefDateFormat);
	time_format = PrefGetPreference(prefTimeFormat);

	TimSecondsToDateTime(seconds, &dt);
	DateToAscii(dt.month, dt.day, dt.year, date_format, buf);
	StrCat(buf, " ");
	TimeToAscii(dt.hour, dt.minute, time_format, &buf[StrLen(buf)]);
	StrPrintF(&buf[StrLen(buf)], " (%d sec)", dt.second);
    }

    set_field(field_id, buf, 0);
}

void print_ratio(char *str, ULong size, ULong compressed)
{
    ULong ratio;

    if (size == 0 || size < compressed) {
	StrCopy(str, "-");
	return;
    }

    ratio = ((size - compressed) * 2000 / size + 1) / 2;
    StrPrintF(str, "%ld.%ld%%", ratio / 10, ratio % 10);
}

static void set_field_editbale(Word field_id, Boolean editable)
{
    FieldPtr field;
    FieldAttrType attr;

    field = GetObjectPtr(field_id);
    FldGetAttributes(field, &attr);
    attr.editable = editable;
    FldSetAttributes(field, &attr);
}

static void update_db_info(void)
{
    char buf[12];
    FormPtr form;
#ifndef UNPZ
    DWord rom_ver;
#endif

    set_field(INFO_NAME_FIELD_ID, db_info.name, 0);
    set_field(INFO_TYPE_FIELD_ID, (char *) &db_info.type, sizeof(ULong));
    set_field(INFO_CREATOR_FIELD_ID, (char *) &db_info.creator, sizeof(ULong));

    set_attrs(db_info.attributes, false);

    set_date_field(INFO_CDAT_FIELD_ID, db_info.creationDate);
    set_date_field(INFO_MDAT_FIELD_ID, db_info.modificationDate);
    set_date_field(INFO_BDAT_FIELD_ID, db_info.lastBackupDate);

    StrPrintF(buf, "%u", db_info.version);
    set_field(INFO_VERS_FIELD_ID, buf, 0);
    StrPrintF(buf, "%lu", db_info.modificationNumber);
    set_field(INFO_MODS_FIELD_ID, buf, 0);

    StrPrintF(buf, "%u", db_info.recordList.numRecords);
    set_field(INFO_RECS_FIELD_ID, buf, 0);
    StrPrintF(buf, "%lu", db_size);
    set_field(INFO_SIZE_FIELD_ID, buf, 0);

    form = FrmGetActiveForm();
    if (db_compressed == 0) {
	if (db_writable) {
	    set_field_editbale(INFO_NAME_FIELD_ID, true);
	    set_field_editbale(INFO_TYPE_FIELD_ID, true);
	    set_field_editbale(INFO_CREATOR_FIELD_ID, true);
	    set_field_editbale(INFO_ATTR_FIELD_ID, true);
	    FrmShowObject(form, FrmGetObjectIndex(form, INFO_APPLY_ID));
	    FrmShowObject(form, FrmGetObjectIndex(form, INFO_DELETE_ID));
#ifndef UNPZ
	    if ((db_info.attributes & dmHdrAttrCopyPrevention) == 0
	     && FtrGet(sysFtrCreator, sysFtrNumROMVersion, &rom_ver) == noErr
	     && sysGetROMVerMajor(rom_ver) >= 3) {
		FrmShowObject(form, FrmGetObjectIndex(form, INFO_BEAM_ID));
	    }
#endif
	}
    } else {
	FrmShowObject(form, FrmGetObjectIndex(form, INFO_RATIO_LABEL_ID));
	FrmShowObject(form, FrmGetObjectIndex(form, INFO_RATIO_FIELD_ID));
	FrmShowObject(form, FrmGetObjectIndex(form, INFO_CMPR_LABEL_ID));
	FrmShowObject(form, FrmGetObjectIndex(form, INFO_CMPR_FIELD_ID));
	print_ratio(buf, db_size, db_compressed);
	set_field(INFO_RATIO_FIELD_ID, buf, 0);
	StrPrintF(buf, "%lu", db_compressed);
	set_field(INFO_CMPR_FIELD_ID, buf, 0);
    }
}


/*
 * modify information
 */

static Boolean parse_attr(CharPtr str, UInt *attr)
{
    char c;

    *attr = 0;

    do {
	c = *str++;
	if (c >= '0' && c <= '9') {
	    *attr = (*attr << 4) | ((c - '0') & 0x0f);
	} else if ((c >= 'A' && c <= 'F') || (c >= 'a' && c <= 'f')) {
	    *attr = (*attr << 4) | ((c - 'A' + 10) & 0x0f);
	} else {
	    return false;
	}
    } while (*str != '\0');

    return true;
}

static void do_attr_dialog(void)
{
    CharPtr attr_str;
    UInt attr;
    FormPtr form;
    int i;

    attr_str = FldGetTextPtr(GetObjectPtr(INFO_ATTR_FIELD_ID));
    if (attr_str != NULL && parse_attr(attr_str, &attr)) {
	attr = (attr & ~dmHdrAttrsShouldNotModify)
		| (db_info.attributes & dmHdrAttrsShouldNotModify);
    } else {
	attr = db_info.attributes;
    }

    form = FrmInitForm(ATTR_DIALOG_ID);
    for (i = 0; i < 16; ++i) {
	if ((attr & (1 << i)) != 0) {
	    FrmSetControlValue(form,
		    FrmGetObjectIndex(form, ATTR_RES_DB_ID + i), true);
	}
	if (! db_writable) {
	    CtlSetEnabled(GetObjectPtrIn(form, ATTR_RES_DB_ID + i), false);
	}
    }
    FrmDoDialog(form);
    attr = 0;
    for (i = 0; i < 16; ++i) {
	if (CtlGetValue(GetObjectPtrIn(form, ATTR_RES_DB_ID + i))) {
	    attr |= (1 << i);
	}
    }
    FrmDeleteForm(form);

    if (db_writable) {
	set_attrs(attr, true);
    }
}

static void modify_db_info(void)
{
    CharPtr name, type, creator, attr_str;
    UInt attr;
    Boolean attr_good = false;

    name = FldGetTextPtr(GetObjectPtr(INFO_NAME_FIELD_ID));
    if (name != NULL
     && (name[0] == '\0' || StrCompare(name, db_info.name) == 0)) {
	name = NULL;
    }

    type = FldGetTextPtr(GetObjectPtr(INFO_TYPE_FIELD_ID));
    if (type != NULL && StrLen(type) != sizeof(ULong)) {
	type = NULL;
    }

    creator = FldGetTextPtr(GetObjectPtr(INFO_CREATOR_FIELD_ID));
    if (creator != NULL && StrLen(creator) != sizeof(ULong)) {
	creator = NULL;
    }

    attr_str = FldGetTextPtr(GetObjectPtr(INFO_ATTR_FIELD_ID));
    if (attr_str != NULL) {
	attr_good = parse_attr(attr_str, &attr);
	attr = (attr & ~dmHdrAttrsShouldNotModify)
		| (db_info.attributes & dmHdrAttrsShouldNotModify);
    }

    DmSetDatabaseInfo(db_spec.card_no, db_spec.db_id, name,
	    attr_good ? &attr : NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL,
	    (ULong *) type, (ULong *) creator);
}


/*
 * delete database
 */

static Boolean delete_db(void)
{
    Word button;
    int err;

    button = FrmCustomAlert(DELETE_ALERT_ID, "database", db_info.name, " ");
    if (button == DELETE_OK_ID) {
	err = DmDeleteDatabase(db_spec.card_no, db_spec.db_id);
	if (err == noErr) {
	    prefs.last_selection.db_id = NULL;
	    return true;
	}
	FrmCustomAlert(ERROR_ALERT_ID, "DmDeleteDatabase() failed", " ", " ");
    }
    return false;
}


/*
 * beam database
 */

#ifndef UNPZ
static Err beam_write(const void *data, ULong *size, void *user_data)
{
    Err err;

    *size = ExgSend((ExgSocketPtr) user_data, data, *size, &err);
    return err;
}

static void beam_db(UInt card_no, LocalID db_id, CharPtr name)
{
    ExgSocketType socket;
    Err err;

    MemSet(&socket, sizeof socket, 0);
    socket.target = BEAM_TARGET_CREATOR;
    socket.count = 1;
    socket.description = name;

    err = ExgPut(&socket);
    if (err == noErr) {
	err = ExgDBWrite(beam_write, &socket, name, db_id, card_no);
	err = ExgDisconnect(&socket, err);
    }
}
#endif


/*
 * event handlers
 */

Boolean handle_info_form_event(EventPtr event)
{
    Boolean handled = false;

    CALLBACK_PROLOGUE

    switch (event->eType) {
    case frmOpenEvent:
	update_db_info();
	FrmDrawForm(FrmGetActiveForm());
	handled = true;
	break;

    case ctlSelectEvent:
	switch (event->data.ctlSelect.controlID) {
	case INFO_ATTR_SELECTOR_ID:
	    do_attr_dialog();
	    handled = true;
	    break;

	case INFO_APPLY_ID:
	    modify_db_info();
	    FrmUpdateForm(prefs.last_form_id, UPDATE_DB_LIST);

	    /* fall through */

	case INFO_DONE_ID:
	done:
	    FrmReturnToForm(prefs.last_form_id);
	    handled = true;
	    break;

	case INFO_DELETE_ID:
	    if (delete_db()) {
		FrmUpdateForm(prefs.last_form_id, UPDATE_DB_LIST);
		goto done;
	    }
	    handled = true;
	    break;

#ifndef UNPZ
	case INFO_BEAM_ID:
	    beam_db(db_spec.card_no, db_spec.db_id, db_info.name);
	    handled = true;
	    break;
#endif
	}
	break;

    default:
	break;
    }

    CALLBACK_EPILOGUE

    return handled;
}
