/* db_list.c */

#include <Pilot.h>

#include "callback.h"

#include "pz.h"


/*
 * constants
 */

#ifndef dmHdrAttrHidden
#define dmHdrAttrHidden 0x0100
#endif

#ifndef dmHdrAttrLaunchableData
#define dmHdrAttrLaunchableData 0x0200
#endif

#ifndef memHeapFlagReadOnly
#define memHeapFlagReadOnly 0x0001
#endif

#define APP_DB_TYPE 'appl'
#define PANEL_DB_TYPE 'panl'
#define HACK_DB_TYPE 'HACK'
#define DA_DB_TYPE 'DAcc'
#define LOCALIZER_CREATOR 'JLoc'
#define DATA_DB_TYPE1 'DATA'
#define DATA_DB_TYPE2 'Data'
#define DOC_DB_TYPE 'TEXt'
#define DOC_CREATOR 'REAd'
#define IMAGE_DB_TYPE 'vIMG'

#define FLASH_HEAP_ID 4

#define DB_RECS_UNKNOWN 0xffffffff
#define DB_SIZE_UNKNOWN 0xffffffff

#define DB_INFO_WIDTH 28


/*
 * global variables
 */

static const char *where_strs[where_other + 1]
	= { "", "-", "rom", "flsh", "ext" };

#ifndef UNPZ
static const char *kind_strs[kind_unknown + 1]
	= { "", "sys", "pz", "app", "pnl", "hac", "da", "loc",
	    "dat", "doc", "img", "?" };
#endif

struct db_rec *db_list;
int num_dbs;


/*
 * generate and free functions
 */

static enum db_where where_of(UInt card_no, LocalID db_id)
{
    if (card_no != 0) {
	return where_other;
    } else if (MemLocalIDKind(db_id) == memIDHandle) {
	return where_ram;
    } else if (MemPtrHeapID(MemLocalIDToPtr(db_id, card_no)) == FLASH_HEAP_ID) {
	return where_flash;
    } else {
	return where_rom;
    }
}

static enum db_kind kind_of(UInt card_no, LocalID db_id)
{
    ULong type, creator;

    DmDatabaseInfo(card_no, db_id, NULL, NULL,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, &type, &creator);
#ifndef UNPZ
    if (creator == 'psys'
     || type == 'exgl' || type == 'extn' || type == 'htal' || type == 'libr'
     || type == 'neti' || type == 'sdrv' || type == 'vdrv') {
	return kind_system;
    } else
#endif
    if (type == PZ_DB_TYPE && creator == PZ_CREATOR) {
	return kind_archive;
#ifndef UNPZ
    } else if (type == APP_DB_TYPE) {
	return kind_application;
    } else if (type == PANEL_DB_TYPE) {
	return kind_panel;
    } else if (type == HACK_DB_TYPE) {
	return kind_hack;
    } else if (type == DA_DB_TYPE) {
	return kind_da;
    } else if (creator == LOCALIZER_CREATOR) {
	return kind_localizer;
    } else if (type == DATA_DB_TYPE1 || type == DATA_DB_TYPE2) {
	return kind_data;
    } else if (type == DOC_DB_TYPE && creator == DOC_CREATOR) {
	return kind_doc;
    } else if (type == IMAGE_DB_TYPE) {
	return kind_image;
#endif
    } else {
	return kind_unknown;
    }
}

void gen_db_list(enum db_where where, enum db_kind kind)
{
    int max_dbs, num_cards, card_dbs, i, j;
    struct db_rec *p;

    num_dbs = 0;

    max_dbs = 0;
    num_cards = MemNumCards();
    for (i = 0; i < num_cards; ++i) {
	max_dbs += DmNumDatabases(i);
    }
    db_list = MemPtrNew(sizeof(struct db_rec) * max_dbs);
    if (db_list == NULL) {
	return;
    }

    for (i = 0; i < num_cards; ++i) {
	card_dbs = DmNumDatabases(i);
	for (j = 0; j < card_dbs; ++j) {
	    p = &db_list[num_dbs];
	    p->spec.card_no = i;
	    p->spec.db_id = DmGetDatabase(i, j);
	    p->where = where_of(p->spec.card_no, p->spec.db_id);
	    if (where == where_all || p->where == where) {
		p->kind = kind_of(p->spec.card_no, p->spec.db_id);
		if (kind == kind_all || p->kind == kind) {
		    p->recs = DB_RECS_UNKNOWN;
		    p->size = DB_SIZE_UNKNOWN;
		    ++num_dbs;
		}
	    }
	}
    }
}

void free_db_list(void)
{
    if (db_list != NULL) {
	MemPtrFree(db_list);
    }

    db_list = NULL;
    num_dbs = 0;
}


/*
 * sort functions
 */

static void get_db_size(struct db_rec *p)
{
    if (p->recs == DB_RECS_UNKNOWN && p->size == DB_SIZE_UNKNOWN) {
	DmDatabaseSize(
		p->spec.card_no, p->spec.db_id, &p->recs, &p->size, NULL);
    }
}

#ifndef UNPZ
static Int compare_ulongs(ULong ul1, ULong ul2)
{
    return (ul1 == ul2) ? 0 : ((ul1 < ul2) ? -1 : 1);
}

static Int compare_types(ULong type1, ULong type2)
{
    Int result;

    result = StrNCaselessCompare(
	    (CharPtr) &type1, (CharPtr) &type2, sizeof(ULong));
    if (result == 0) {
	result = compare_ulongs(type1, type2);
    }
    return result;
}
#endif

static Int compare(struct db_rec *p1, struct db_rec *p2, Long sort_by)
{
    char name1[dmDBNameLength], name2[dmDBNameLength];
#ifndef UNPZ
    UInt attr1, attr2, vers1, vers2;
    ULong cdat1, mdat1, bdat1, mods1, type1, creator1;
    ULong cdat2, mdat2, bdat2, mods2, type2, creator2;
#endif
    Int result;

    CALLBACK_PROLOGUE

#ifndef UNPZ
    switch (sort_by) {
    case sort_by_where:
	result = (Int) p1->where - (Int) p2->where;
	break;

    case sort_by_kind:
	result = (Int) p1->kind - (Int) p2->kind;
	break;

    case sort_by_recs:
	get_db_size(p1);
	get_db_size(p2);
	result = - compare_ulongs(p1->recs, p2->recs);
	break;

    case sort_by_size:
	get_db_size(p1);
	get_db_size(p2);
	result = - compare_ulongs(p1->size, p2->size);
	break;

    default:
	result = 0;
	break;
    }

    if (result == 0) {
	DmDatabaseInfo(p1->spec.card_no, p1->spec.db_id, name1, &attr1, &vers1,
		&cdat1, &mdat1, &bdat1, &mods1, NULL, NULL, &type1, &creator1);
	DmDatabaseInfo(p2->spec.card_no, p2->spec.db_id, name2, &attr2, &vers2,
		&cdat2, &mdat2, &bdat2, &mods2, NULL, NULL, &type2, &creator2);

	switch (sort_by) {
	case sort_by_type:
	    result = compare_types(type1, type2);
	    break;

	case sort_by_creator:
	    result = compare_types(creator1, creator2);
	    break;

	case sort_by_attr:
	    result = compare_ulongs(attr1, attr2);
	    break;

	case sort_by_cdat:
	    result = - compare_ulongs(cdat1, cdat2);
	    break;

	case sort_by_mdat:
	    result = - compare_ulongs(mdat1, mdat2);
	    break;

	case sort_by_bdat:
	    result = - compare_ulongs(bdat1, bdat2);
	    break;

	case sort_by_vers:
	    result = - compare_ulongs(vers1, vers2);
	    break;

	case sort_by_mods:
	    result = - compare_ulongs(mods1, mods2);
	    break;

	default:
	    break;
	}

	if (result == 0) {
	    result = StrCaselessCompare(name1, name2);
	}
    }
#else
    DmDatabaseInfo(p1->spec.card_no, p1->spec.db_id, name1,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
    DmDatabaseInfo(p2->spec.card_no, p2->spec.db_id, name2,
	    NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);

    result = StrCaselessCompare(name1, name2);
#endif

    CALLBACK_EPILOGUE

    return result;
}

void sort_db_list(enum db_sort_by sort_by)
{
    SysQSort(db_list, num_dbs,
	    sizeof(struct db_rec), (CmpFuncPtr) compare, sort_by);
}


/*
 * search functions
 */

static int find_fast(const char *str)
{
    int min, max, mid;
    struct db_rec *p;
    char name[dmDBNameLength];
    ULong type, creator;

    min = 0;
    max = num_dbs - 1;

    while (min <= max) {
	mid = (min + max + 1) / 2;

	/* (min < mid <= max) or (min == mid == max) */

	p = &db_list[mid];
	DmDatabaseInfo(p->spec.card_no, p->spec.db_id, name, NULL,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, &type, &creator);
#ifndef UNPZ
	if (prefs.sort_by == sort_by_type) {
	    MemMove(name, (char *) &type, sizeof(ULong));
	    name[sizeof(ULong)] = '\0';
	} else if (prefs.sort_by == sort_by_creator) {
	    MemMove(name, (char *) &creator, sizeof(ULong));
	    name[sizeof(ULong)] = '\0';
	}
#endif

	if (StrCaselessCompare(name, str) >= 0) {
	    max = mid - 1;
	} else if (min == mid) {	/* don't ask me why it works... */
	    min = mid + 1;
	} else {
	    min = mid;
	}
    }

    return min;
}

#ifndef UNPZ
static int find_slow(const char *str)
{
    int i;
    struct db_rec *p;
    char name[dmDBNameLength];

    for (i = 0; i < num_dbs; ++i) {
	p = &db_list[i];
	DmDatabaseInfo(p->spec.card_no, p->spec.db_id, name, NULL,
		NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	if (StrNCaselessCompare(name, str, 1) == 0) {
	    return i;
	}
    }
    return -1;
}
#endif

int find_db(char chr)
{
    char buf[2];

    if ((chr >= '0' && chr <= '9')
     || (chr >= 'A' && chr <= 'Z')
     || (chr >= 'a' && chr <= 'z')) {
	buf[0] = chr;
	buf[1] = '\0';
#ifndef UNPZ
	if (prefs.sort_by == sort_by_name
	 || prefs.sort_by == sort_by_type
	 || prefs.sort_by == sort_by_creator) {
#endif
	    return find_fast(buf);
#ifndef UNPZ
	} else {
	    return find_slow(buf);
	}
#endif
    } else {
	return -1;
    }
}


/*
 * custom draw procedure
 */

void draw_chars(const char *str, int length, int x, RectanglePtr bounds)
{
    if (length == 0) {
	length = StrLen(str);
    }
    WinDrawChars(str, length, bounds->topLeft.x + x, bounds->topLeft.y);
}

void draw_fit_chars(
	const char *str, int x, int width, RectanglePtr bounds, Boolean center)
{
    Int new_width, length;
    Boolean fits;

    new_width = width;
    length = StrLen(str);
    FntCharsInWidth(str, &new_width, &length, &fits);
    if (fits) {
	if (center) {
	    x += (width - new_width) / 2;
	}
	draw_chars(str, length, x, bounds);
    } else {
	new_width = width - FntCharWidth('.') * 3;
	FntCharsInWidth(str, &new_width, &length, &fits);
	draw_chars(str, length, x, bounds);
	draw_chars("...", 3, x + new_width, bounds);
    }
}

#ifndef UNPZ
static void print_type_creator(char *str, ULong type_creator)
{
    MemMove(str, &type_creator, sizeof(ULong));
    str[sizeof(ULong)] = '\0';
}
#endif

void print_size(char *str, ULong size)
{
    if (size >= 102349) {
	StrPrintF(str, "%ldk", (size + 512) / 1024);
    } else {
	size = (size * 10 + 512) / 1024;
	StrPrintF(str, "%ld.%ldk", size / 10, size % 10);
    }
}

#ifndef UNPZ
static void print_attr(char *str, UInt attr)
{
    str[0] = (attr & dmHdrAttrOpen) ? 'o' :
	    ((attr & dmHdrAttrLaunchableData) ? 'L' :
	    ((attr & dmHdrAttrHidden) ? 'h' :
	    ((attr & dmHdrAttrStream) ? 's' : '-')));
    str[1] = (attr & dmHdrAttrCopyPrevention) ? 'c' :
	    ((attr & dmHdrAttrResetAfterInstall) ? 'r' :
	    ((attr & dmHdrAttrOKToInstallNewer) ? 'n' : '-'));
    str[2] = (attr & dmHdrAttrBackup) ? 'b' : '-';
    str[3] = (attr & dmHdrAttrReadOnly) ? 'R' :
	    ((attr & dmHdrAttrAppInfoDirty) ? 'D' : '-');
    str[4] = (attr & dmHdrAttrResDB) ? 'r' : 'd';
    str[5] = '\0';
}

static void print_date(char *str, ULong date)
{
    DateTimeType dt;
    SWord this_year;
    DateFormatType date_format;
    char date_separator;

    if (date == 0) {
	StrCopy(str, "-");
	return;
    }

    TimSecondsToDateTime(TimGetSeconds(), &dt);
    this_year = dt.year;
    TimSecondsToDateTime(date, &dt);
    if (dt.year != this_year) {
	StrPrintF(str, "%dy%02d", dt.month, dt.year % 100);
	return;
    }

    date_format = PrefGetPreference(prefDateFormat);

    switch (date_format) {
    case dfDMYWithDots:
    case dfYMDWithDots:
	date_separator = '.';
	break;

    case dfDMYWithDashes:
    case dfYMDWithDashes:
	date_separator = '-';
	break;

    default:
	date_separator = '/';
	break;
    }

    switch (date_format) {
    case dfDMYWithSlashes:
    case dfDMYWithDots:
    case dfDMYWithDashes:
	StrPrintF(str, "%d%c%d", dt.day, date_separator, dt.month);
	break;

    default:
	StrPrintF(str, "%d%c%d", dt.month, date_separator, dt.day);
	break;
    }
}
#endif

void draw_db_item(VoidPtr table, Word row, Word column, RectanglePtr bounds)
{
    int index;
    struct db_rec *p;
    char name[dmDBNameLength];
    UInt attr, vers;
    ULong cdat, mdat, bdat, mods, type, creator;
    Int x, info_width;
    enum db_info info;

    CALLBACK_PROLOGUE

    WinEraseRectangle(bounds, 0);

    index = get_table_top() + row;

    if (index < num_dbs) {
	p = &db_list[index];

#ifndef UNPZ
	if (prefs.info0 == info0_where) {
#endif
	    draw_chars(where_strs[p->where], 0, 1, bounds);
	    x = 20;
#ifndef UNPZ
	} else if (prefs.info0 == info0_kind) {
	    draw_chars(kind_strs[p->kind], 0, 1, bounds);
	    x = 20;
	} else {
	    x = 2;
	}
#endif

	DmDatabaseInfo(p->spec.card_no, p->spec.db_id, name, &attr,
		&vers, &cdat, &mdat, &bdat, &mods, NULL, NULL, &type, &creator);

#ifndef UNPZ
	if (prefs.info1 != info_none) {
#endif
	    info_width = DB_INFO_WIDTH * 2 + 2;
#ifndef UNPZ
	} else if (prefs.info2 != info_none) {
	    info_width = DB_INFO_WIDTH + 2;
	} else {
	    info_width = 0;
	}
#endif

	draw_fit_chars(
		name, x, bounds->extent.x - x - info_width, bounds, false);

#ifndef UNPZ
	if (prefs.info1 == info_recs || prefs.info1 == info_size
	 || prefs.info2 == info_recs || prefs.info2 == info_size) {
#endif
	    get_db_size(p);
#ifndef UNPZ
	}
#endif

#ifndef UNPZ
	for (x = DB_INFO_WIDTH * 2, info = prefs.info1;
		x > 0; x -= DB_INFO_WIDTH, info = prefs.info2) {
#else
	for (x = DB_INFO_WIDTH * 2, info = info_recs;
		x > 0; x -= DB_INFO_WIDTH, info = info_size) {
#endif
	    switch (info) {
#ifndef UNPZ
	    case info_type:
		print_type_creator(name, type);
		break;

	    case info_creator:
		print_type_creator(name, creator);
		break;
#endif

	    case info_recs:
		StrPrintF(name, "%ld", p->recs);
		break;

	    case info_size:
		print_size(name, p->size);
		break;

#ifndef UNPZ
	    case info_attr:
		print_attr(name, attr);
		break;

	    case info_cdat:
		print_date(name, cdat);
		break;

	    case info_mdat:
		print_date(name, mdat);
		break;

	    case info_bdat:
		print_date(name, bdat);
		break;

	    case info_vers:
		StrPrintF(name, "%d", vers);
		break;

	    case info_mods:
		StrPrintF(name, "%ld", mods);
		break;
#endif

	    default:
		name[0] = '\0';
		break;
	    }

	    if (name[0] != '\0') {
		draw_chars(name, 0, bounds->extent.x - x, bounds);
	    }
	}
    }

    CALLBACK_EPILOGUE
}


/*
 * drawing free memory
 */

void draw_free_memory(Word gadget_id)
{
    RectangleType bounds;
    UInt num_heaps, i, heap_id;
    ULong free, max, total;
    char buf[16];

    GetObjectBounds(gadget_id, &bounds);
    WinEraseRectangle(&bounds, 0);

    total = 0;
    num_heaps = MemNumHeaps(DEFAULT_CARD_NO);
    for (i = 0; i < num_heaps; ++i) {
	heap_id = MemHeapID(DEFAULT_CARD_NO, i);
	if (MemHeapDynamic(heap_id) == false
	 && (MemHeapFlags(heap_id) & memHeapFlagReadOnly) == 0) {
	    MemHeapFreeBytes(heap_id, &free, &max);
	    total += free;
	}
    }

    print_size(buf, total);
    StrCat(buf, " free");
    draw_fit_chars(buf, 0, bounds.extent.x, &bounds, true);
}
