/* stream.c */

#include <Pilot.h>

#include "pz.h"


/*
 * constants
 */

/* from MemoryPrv.h */
#ifndef memHeapFlagReadOnly
#define memHeapFlagReadOnly 0x0001	/* heap is read-only (ROM based) */
#endif

#define STREAM_CONTINUE 0xffffffff

#define STREAM_RECORD_MIN 2048
#define STREAM_RECORD_SIZE 32768


/*
 * stream_info functions
 */

#define CHECK_SKIP \
    if (skip + STREAM_TRAILER_SIZE >= size) { \
	return -1;	/* data too short */ \
    }

#define UPDATE_SKIP(delta) \
    skip += (delta); \
    CHECK_SKIP

static int check_header(
	const struct pz_header *p, ULong size, struct stream_info *stream)
{
    ULong skip = sizeof(struct pz_header);

    CHECK_SKIP

    stream->method = p->cm;
    stream->flag = STREAM_FLAG_UNKNOWN;
    stream->name_offset = 0;
    stream->name_length = 0;

    if (p->id1 != PZ_ID1 || p->id2 != PZ_ID2) {
	return -1;	/* invalid magic number */
    }

    if (p->cm != PZ_CM_DEFLATED) {
	return -1;	/* unknown compression method */
    }

    if (p->flg & PZ_FLG_FEXTRA) {
	Byte *extra = (Byte *) p + skip;

	UPDATE_SKIP(PZ_XLEN_SIZE + extra[0] + extra[1] * 256)
    }

    if (p->flg & PZ_FLG_FNAME) {
	char *name = (char *) p + skip;
	int len = StrLen(name);

	stream->name_offset = skip;
	stream->name_length = len;

	UPDATE_SKIP(len + 1)

	if (StrCaselessCompare(&name[len - DB_EXT_LENGTH], DATA_DB_EXT) == 0
	 || StrCaselessCompare(&name[len - DB_EXT_LENGTH], RES_DB_EXT) == 0) {
	    stream->flag = STREAM_FLAG_VALID;
	}
    }

    if (p->flg & PZ_FLG_FCOMMENT) {
	UPDATE_SKIP(StrLen((char *) p + skip) + 1)
    }

    if (p->flg & PZ_FLG_FHCRC) {
	UPDATE_SKIP(PZ_CRC16_SIZE)
    }

    stream->data_offset = skip;

    return noErr;
}

#undef CHECK_SKIP
#undef UPDATE_SKIP

int count_streams(DmOpenRef db, struct stream_info **streams, int *bad)
{
    int num_records, num_streams, i, err;
    ULong size, uncompressed;
    VoidHand h;
    VoidPtr p;

    num_records = DmNumRecords(db);
    *streams = MemPtrNew(sizeof(struct stream_info) * num_records);
    if (*streams == NULL) {
	*bad = -1;
	return 0;
    }

    *bad = 0;
    num_streams = 0;
    uncompressed = 0;
    for (i = 0; i < num_records; ++i) {
	h = DmQueryRecord(db, i);
	size = MemHandleSize(h);
	p = MemHandleLock(h);
	if ((long) uncompressed >= 0) {
	    err = check_header(p, size, &(*streams)[num_streams]);
	    (*streams)[num_streams].index = i;
	    (*streams)[num_streams].num_records = 1;
	    (*streams)[num_streams].size = size - STREAM_TRAILER_SIZE;
	} else {
	    err = (size <= STREAM_TRAILER_SIZE) ? -1 : noErr;
	    (*streams)[num_streams].num_records += 1;
	    (*streams)[num_streams].size += size - STREAM_TRAILER_SIZE;
	}
	if (err == noErr) {
	    uncompressed = ((ULong) ((Byte *) p)[size - 1] << 24)
			 + ((ULong) ((Byte *) p)[size - 2] << 16)
			 + ((ULong) ((Byte *) p)[size - 3] << 8)
			 + ((ULong) ((Byte *) p)[size - 4]);
	    (*streams)[num_streams].uncompressed = uncompressed;
	    if ((long) uncompressed >= 0) {
		++num_streams;
	    }
	} else {
	    uncompressed = 0;
	    *bad += (*streams)[num_streams].num_records;
	}
	MemHandleUnlock(h);
    }

    if ((long) uncompressed < 0) {
	++*bad;
    }

    if (num_streams == 0) {
	MemPtrFree(*streams);
	*streams = NULL;
    }

    return num_streams;
}

#ifndef UNPZ

int remove_stream(DmOpenRef db, UInt index, UInt num_records)
{
    int i, err;

    for (i = 0; i < num_records; ++i) {
	err = DmRemoveRecord(db, index);
	if (err != noErr) { return err; }
    }

    return noErr;
}

#endif /* ! UNPZ */

void print_stream_name(
	DmOpenRef db, struct stream_info *stream, char *name, int length)
{
    VoidHand h;
    CharPtr p;

    if (stream->name_offset == 0 || stream->name_length == 0) {
	StrCopy(name, "(no name)");
	return;
    }

    h = DmQueryRecord(db, stream->index);
    p = MemHandleLock(h);
    if (stream->name_length < length) {
	MemMove(name, p + stream->name_offset, stream->name_length);
	name[stream->name_length] = '\0';
    } else {
	MemMove(name, p + stream->name_offset, length - 4);
	StrCopy(&name[length - 4], "...");
    }
    MemHandleUnlock(h);
}


/*
 * in_stream functions
 */

static Byte *get_record(struct stream *in)
{
    in->cur_handle = DmQueryRecord(in->db, in->cur_index);
    in->cur_ptr = MemHandleLock(in->cur_handle);

    in->cur_size = MemHandleSize(in->cur_handle)
	    - in->cur_offset - STREAM_TRAILER_SIZE;

    return (Byte *) in->cur_ptr + in->cur_offset;
}

static void unlock_record(struct stream *in)
{
    if (in->cur_handle != NULL) {
	MemHandleUnlock(in->cur_handle);
	in->cur_handle = NULL;
	in->cur_start += in->cur_size;
    }
}

Byte *open_in_stream(
	struct stream *in, DmOpenRef db, struct stream_info *stream)
{
    in->db = db;
    in->first_index = stream->index;
    in->num_records = stream->num_records;

    in->cur_index = in->first_index;
    in->cur_start = 0;
    in->cur_offset = stream->data_offset;

    return get_record(in);
}

void close_in_stream(struct stream *in)
{
    unlock_record(in);
}

ULong avail_in_stream(struct stream *in, ULong total_in, Byte **next_in)
{
    ULong avail_in;

    avail_in = in->cur_start + in->cur_size - total_in;

    if (avail_in == 0) {
	unlock_record(in);

	++in->cur_index;
	if (in->cur_index < in->first_index + in->num_records) {
	    in->cur_offset = 0;
	    *next_in = get_record(in);
	    avail_in = in->cur_size;
	}
    }

    return avail_in;
}


/*
 * out_stream functions
 */

#ifndef UNPZ

static ULong max_free(DmOpenRef db)
{
    UInt card_no, num_heaps, i, heap_id;
    ULong free, dummy, max;
    int err;

    err = DmOpenDatabaseInfo(db, NULL, NULL, NULL, &card_no, NULL);
    if (err != noErr) {
	return 0;
    }

    free = 0;
    num_heaps = MemNumHeaps(card_no);
    for (i = 0; i < num_heaps; ++i) {
	heap_id = MemHeapID(card_no, i);
	if ((MemHeapFlags(heap_id) & memHeapFlagReadOnly) == 0
	 && MemHeapDynamic(heap_id) == false) {
	    err = MemHeapFreeBytes(heap_id, &dummy, &max);
	    if (err == noErr && free < max) {
		free = max;
	    }
	}
    }

    return free;
}

static int create_record(struct stream *out, ULong size)
{
    ULong free;

    out->cur_handle = DmNewRecord(out->db, &out->cur_index, size);
    if (out->cur_handle == NULL) {
	free = max_free(out->db);
	if (free < STREAM_RECORD_MIN + STREAM_RECORD_MIN) {
	    return STORAGE_ERROR;
	}
	free -= STREAM_RECORD_MIN;
	if (free > STREAM_RECORD_SIZE) {
	    free = STREAM_RECORD_SIZE;
	}
	out->cur_handle = DmNewRecord(out->db, &out->cur_index, free);
	if (out->cur_handle == NULL) {
	    return STORAGE_ERROR;
	}
    }

    ++out->num_records;
    if (out->num_records == 1) {
	out->first_index = out->cur_index;
    }

    out->cur_ptr = MemHandleLock(out->cur_handle);
    out->cur_size = MemHandleSize(out->cur_handle) - STREAM_TRAILER_SIZE;
    out->cur_offset = 0;

    return noErr;
}

static void release_record(struct stream *out, ULong trailer)
{
    Byte buf[STREAM_TRAILER_SIZE];

    if (out->cur_handle != NULL) {
	buf[0] = trailer;
	buf[1] = trailer >> 8;
	buf[2] = trailer >> 16;
	buf[3] = trailer >> 24;
	DmWrite(out->cur_ptr, out->cur_offset, buf, STREAM_TRAILER_SIZE);

	MemHandleUnlock(out->cur_handle);
	if (out->cur_offset < out->cur_size) {
	    MemHandleResize(out->cur_handle,
		    out->cur_offset + STREAM_TRAILER_SIZE);
	}
	out->cur_handle = NULL;

	DmReleaseRecord(out->db, out->cur_index, true);

	out->cur_start += out->cur_offset;
    }
}

int open_out_stream(
	struct stream *out, DmOpenRef db, UInt index, ULong max_size)
{
    out->db = db;
    out->first_index = index;
    out->num_records = 0;

    out->cur_index = out->first_index;
    out->cur_start = 0;

    return create_record(out, min(max_size, STREAM_RECORD_SIZE));
}

void close_out_stream(struct stream *out, ULong trailer)
{
    release_record(out, trailer);
}

int write_out_stream(struct stream *out, const void *ptr, ULong size)
{
    ULong some;
    int err = noErr;

    while (size > 0) {
	some = min(size, out->cur_size - out->cur_offset);
	if (some > 0) {
	    DmWrite(out->cur_ptr, out->cur_offset, ptr, some);
	    out->cur_offset += some;
	    ptr = (Byte *) ptr + some;
	    size -= some;
	}

	if (size > 0) {
	    release_record(out, STREAM_CONTINUE);

	    ++out->cur_index;
	    err = create_record(out, STREAM_RECORD_SIZE);
	    if (err != noErr) { break; }
	}
    }

    return err;
}

#endif /* ! UNPZ */
